(* ========================================================================= *)
(* Automated support for general recursive definitions.                      *)
(*                                                                           *)
(*              (c) Copyright, John Harrison 1998-2007                       *)
(* ========================================================================= *)

let CASEWISE_DEF = new_recursive_definition list_RECURSION
 `(CASEWISE [] f x = @y. T) /\
  (CASEWISE (CONS h t) f x =
        if ?y. FST h y = x then SND h f (@y. FST h y = x)
        else CASEWISE t f x)`;;

let CASEWISE = prove
 (`(CASEWISE [] f x = @y. T) /\
   (CASEWISE (CONS (s,t) clauses) f x =
        if ?y. s y = x then t f (@y. s y = x) else CASEWISE clauses f x)`,
  REWRITE_TAC[CASEWISE_DEF]);;

(* ------------------------------------------------------------------------- *)
(* Conditions for all the clauses in a casewise definition to hold.          *)
(* ------------------------------------------------------------------------- *)

let CASEWISE_CASES = prove
 (`!clauses c x.
    (?s t a. MEM (s,t) clauses /\ (s a = x) /\
             (CASEWISE clauses c x = t c a)) \/
    ~(?s t a. MEM (s,t) clauses /\ (s a = x)) /\
    (CASEWISE clauses c x = @y. T)`,
  MATCH_MP_TAC list_INDUCT THEN
  REWRITE_TAC[MEM; CASEWISE; FORALL_PAIR_THM; PAIR_EQ] THEN
  REPEAT STRIP_TAC THEN COND_CASES_TAC THEN ASM_MESON_TAC[]);;

let CASEWISE_WORKS = prove
 (`!clauses c:C.
     (!s t s' t' x y. MEM (s,t) clauses /\ MEM (s',t') clauses /\ (s x = s' y)
                      ==> (t c x = t' c y))
     ==> ALL (\(s:P->A,t). !x. CASEWISE clauses c (s x) :B = t c x) clauses`,
  REWRITE_TAC[GSYM ALL_MEM; FORALL_PAIR_THM] THEN
  MESON_TAC[CASEWISE_CASES]);;

(* ------------------------------------------------------------------------- *)
(* Various notions of admissibility, with tail recursion and preconditions.  *)
(* ------------------------------------------------------------------------- *)

let admissible = new_definition
 `admissible(<<) p s t <=>
        !f g a. p f a /\ p g a /\ (!z. z << s(a) ==> (f z = g z))
                ==> (t f a = t g a)`;;

let tailadmissible = new_definition
 `tailadmissible(<<) p s t <=>
        ?P G H. (!f a y. P f a /\ y << G f a ==> y << s a) /\
                (!f g a. (!z. z << s(a) ==> (f z = g z))
                         ==> (P f a = P g a) /\
                             (G f a = G g a) /\ (H f a = H g a)) /\
                (!f a:P. p f a ==> (t (f:A->B) a =
                                    if P f a then f(G f a) else H f a))`;;

let superadmissible = new_definition
 `superadmissible(<<) p s t <=>
        admissible(<<) (\f a. T) s p ==> tailadmissible(<<) p s t`;;

(* ------------------------------------------------------------------------- *)
(* Combinators for "superadmissible".                                        *)
(* ------------------------------------------------------------------------- *)

let SUPERADMISSIBLE_COND = prove
 (`!(<<) p P s h k:(A->B)->P->B.
        admissible(<<) p s P /\
        superadmissible(<<) (\f x. p f x /\ P f x) s h /\
        superadmissible(<<) (\f x. p f x /\ ~P f x) s k
        ==> superadmissible(<<) p s (\f x. if P f x then h f x else k f x)`,
  REWRITE_TAC[superadmissible; admissible] THEN REPEAT GEN_TAC THEN
  DISCH_THEN(CONJUNCTS_THEN2 ASSUME_TAC MP_TAC) THEN
  DISCH_THEN(fun th -> DISCH_TAC THEN CONJUNCTS_THEN MP_TAC th) THEN
  ANTS_TAC THENL [ASM_MESON_TAC[]; ALL_TAC] THEN
  DISCH_THEN(fun th -> ANTS_TAC THENL [ASM_MESON_TAC[]; MP_TAC th]) THEN
  REWRITE_TAC[tailadmissible] THEN
  REWRITE_TAC[LEFT_IMP_EXISTS_THM; RIGHT_IMP_FORALL_THM] THEN
  MAP_EVERY X_GEN_TAC
   [`P1:(A->B)->P->bool`; `G1:(A->B)->P->A`; `H1:(A->B)->P->B`;
    `P2:(A->B)->P->bool`; `G2:(A->B)->P->A`; `H2:(A->B)->P->B`] THEN
  REWRITE_TAC[TAUT `(a1 /\ b1 /\ c1 ==> a2 /\ b2 /\ c2 ==> x) <=>
                    (a1 /\ a2) /\ (b1 /\ b2) /\ (c1 /\ c2) ==> x`] THEN
  DISCH_THEN(fun th ->
   MAP_EVERY EXISTS_TAC
   [`\f:A->B a:P. if p f a then if P f a then P2 f a else P1 f a else F`;
   `\f:A->B a:P. if p f a then if P f a then G2 f a else G1 f a else z:A`;
   `\f:A->B a:P. if p f a then if P f a then H2 f a else H1 f a else w:B`] THEN
   MP_TAC th) THEN
  REWRITE_TAC[] THEN REPEAT(MATCH_MP_TAC MONO_AND THEN CONJ_TAC) THENL
   [ASM_MESON_TAC[];
    POP_ASSUM_LIST(MP_TAC o end_itlist CONJ);
    ALL_TAC] THEN
  REWRITE_TAC[IMP_IMP; AND_FORALL_THM] THEN
  REPEAT(MATCH_MP_TAC MONO_FORALL THEN GEN_TAC) THEN
  DISCH_THEN(fun th -> DISCH_TAC THEN MP_TAC th) THEN
  ASM_REWRITE_TAC[] THEN ASM_MESON_TAC[]);;

let ADMISSIBLE_IMP_SUPERADMISSIBLE = prove
 (`!(<<) p s t:(A->B)->P->B.
      admissible(<<) p s t ==> superadmissible(<<) p s t`,
  REWRITE_TAC[admissible; superadmissible; tailadmissible] THEN
  REPEAT STRIP_TAC THEN
  MAP_EVERY EXISTS_TAC
   [`\f:A->B x:P. F`;
    `\f:A->B. (anything:P->A)`;
    `\f:A->B a:P. if p f a then t f a :B else fixed`] THEN
  ASM_REWRITE_TAC[] THEN ASM_MESON_TAC[]);;

let TAIL_IMP_SUPERADMISSIBLE = prove
 (`!(<<) p s t:(A->B)->P->A.
      admissible(<<) p s t /\
      (!f a. p f a ==> !y. y << t f a ==> y << s a)
      ==> superadmissible(<<) p s (\f x. f(t f x))`,
  REWRITE_TAC[admissible; superadmissible; tailadmissible] THEN
  REPEAT STRIP_TAC THEN MAP_EVERY EXISTS_TAC
   [`\f:A->B x:P. T`;
    `\f:A->B a:P. if p f a then t f a :A else s a`;
    `\f:A->B. anything:P->B`] THEN
  ASM_REWRITE_TAC[] THEN ASM_MESON_TAC[]);;

(* ------------------------------------------------------------------------- *)
(* Analogous combinators for pure admissibility.                             *)
(* ------------------------------------------------------------------------- *)

let ADMISSIBLE_COND = prove
 (`!(<<) p P s h k:(A->B)->P->B.
        admissible(<<) p s P /\
        admissible(<<) (\f x. p f x /\ P f x) s h /\
        admissible(<<) (\f x. p f x /\ ~P f x) s k
        ==> admissible(<<) p s (\f x. if P f x then h f x else k f x)`,
  REPEAT GEN_TAC THEN
  REWRITE_TAC[admissible; AND_FORALL_THM] THEN
  REPEAT(MATCH_MP_TAC MONO_FORALL THEN GEN_TAC) THEN
  DISCH_THEN(fun th -> STRIP_TAC THEN MP_TAC th) THEN
  ASM_REWRITE_TAC[] THEN ASM_MESON_TAC[]);;

let ADMISSIBLE_CONST = prove
 (`admissible(<<) p s (\f. c)`,
  REWRITE_TAC[admissible]);;

let ADMISSIBLE_COMB = prove
 (`!(<<) p s:P->A g:(A->B)->P->C->D y:(A->B)->P->C.
        admissible(<<) p s g /\ admissible(<<) p s y
        ==> admissible(<<) p s (\f x. (g f x) (y f x))`,
  SIMP_TAC[admissible] THEN MESON_TAC[]);;

let ADMISSIBLE_BASE = prove
 (`!(<<) p s t.
        (!f a. p f a ==> t a << s a)
        ==> admissible((<<):A->A->bool) p s (\f:A->B x:P. f(t x))`,
  REWRITE_TAC[admissible] THEN MESON_TAC[]);;

let ADMISSIBLE_NEST = prove
 (`!(<<) p s t.
        admissible(<<) p s t /\
        (!f a. p f a ==> t f a << s a)
        ==> admissible((<<):A->A->bool) p s (\f:A->B x:P. f(t f x))`,
  REWRITE_TAC[admissible] THEN MESON_TAC[]);;

let ADMISSIBLE_LAMBDA = prove
 (`!(<<) p s:P->A t:C->(A->B)->P->D.
   (!u. admissible(<<) p s (\f x. t u f x))
   ==> admissible(<<) p s (\f x u. t u f x)`,
  REWRITE_TAC[admissible; FUN_EQ_THM] THEN MESON_TAC[]);;

(* ------------------------------------------------------------------------- *)
(* Add sums into the mix; it's occasionally useful.                          *)
(* ------------------------------------------------------------------------- *)

let ADMISSIBLE_NSUM = prove
 (`!(<<) p:(B->C)->P->bool s:P->A h a b.
        (!k. admissible(<<) (\f x. a(x) <= k /\ k <= b(x) /\ p f x)
                            s (\f x. h f x k))
        ==> admissible(<<) p s (\f x. nsum(a(x)..b(x)) (h f x))`,
  REWRITE_TAC[admissible] THEN REPEAT STRIP_TAC THEN
  MATCH_MP_TAC NSUM_EQ_NUMSEG THEN ASM_MESON_TAC[]);;

let ADMISSIBLE_SUM = prove
 (`!(<<) p:(B->C)->P->bool s:P->A h a b.
        (!k. admissible(<<) (\f x. a(x) <= k /\ k <= b(x) /\ p f x)
                            s (\f x. h f x k))
   ==> admissible(<<) p s (\f x. sum(a(x)..b(x)) (h f x))`,
  REWRITE_TAC[admissible] THEN REPEAT STRIP_TAC THEN
  MATCH_MP_TAC SUM_EQ_NUMSEG THEN ASM_MESON_TAC[]);;

(* ------------------------------------------------------------------------- *)
(* Combine general WF/tail recursion theorem with casewise definitions.      *)
(* ------------------------------------------------------------------------- *)

let WF_REC_TAIL_GENERAL' = prove
 (`!P G H H'.
         WF (<<) /\
         (!f g x. (!z. z << x ==> (f z = g z))
                  ==> (P f x <=> P g x) /\
                      (G f x = G g x) /\ (H' f x = H' g x)) /\
         (!f x y. P f x /\ y << G f x ==> y << x) /\
         (!f x. H f x = if P f x then f(G f x) else H' f x)
         ==> ?f. !x. f x = H f x`,
  REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN
  MATCH_MP_TAC WF_REC_TAIL_GENERAL THEN ASM_MESON_TAC[]);;

let WF_REC_CASES = prove
 (`!(<<) clauses.
        WF((<<):A->A->bool) /\
        ALL (\(s,t). ?P G H.
                     (!f a y. P f a /\ y << G f a ==> y << s a) /\
                     (!f g a. (!z. z << s(a) ==> (f z = g z))
                              ==> (P f a = P g a) /\
                                  (G f a = G g a) /\ (H f a = H g a)) /\
                     (!f a:P. t f a = if P f a then f(G f a) else H f a))
            clauses
        ==> ?f:A->B. !x. f x = CASEWISE clauses f x`,
  REPEAT STRIP_TAC THEN MATCH_MP_TAC WF_REC_TAIL_GENERAL' THEN
  FIRST_X_ASSUM(MP_TAC o check(is_binary "ALL" o concl)) THEN
  SPEC_TAC(`clauses:((P->A)#((A->B)->P->B))list`,
           `clauses:((P->A)#((A->B)->P->B))list`) THEN
  ASM_REWRITE_TAC[] THEN POP_ASSUM(K ALL_TAC) THEN
  MATCH_MP_TAC list_INDUCT THEN
  REWRITE_TAC[ALL; CASEWISE; FORALL_PAIR_THM] THEN CONJ_TAC THENL
   [MAP_EVERY EXISTS_TAC
     [`\f:A->B x:A. F`; `\f:A->B. anything:A->A`; `\f:A->B x:A. @y:B. T`] THEN
    REWRITE_TAC[];
    ALL_TAC] THEN
  MAP_EVERY X_GEN_TAC
    [`s:P->A`; `t:(A->B)->P->B`; `clauses:((P->A)#((A->B)->P->B))list`] THEN
  DISCH_THEN(fun th -> DISCH_THEN(CONJUNCTS_THEN2 MP_TAC ASSUME_TAC) THEN
                       MP_TAC th) THEN
  ASM_REWRITE_TAC[] THEN POP_ASSUM_LIST(K ALL_TAC) THEN
  REWRITE_TAC[LEFT_IMP_EXISTS_THM] THEN REWRITE_TAC[RIGHT_IMP_FORALL_THM] THEN
  MAP_EVERY X_GEN_TAC
   [`P1:(A->B)->A->bool`; `G1:(A->B)->A->A`; `H1:(A->B)->A->B`;
    `P2:(A->B)->P->bool`; `G2:(A->B)->P->A`; `H2:(A->B)->P->B`] THEN
  REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN
  EXISTS_TAC
   `\f:A->B x:A. if ?y:P. s y = x then P2 f (@y. s y = x) else P1 f x:bool` THEN
  EXISTS_TAC `\f:A->B x:A.
     if ?y:P. s y = x then G2 f (@y. s y = x) else G1 f x:A` THEN
  EXISTS_TAC `\f:A->B x:A. if ?y:P. s y = x
                           then H2 f (@y. s y = x) else H1 f x:B` THEN
  ASM_MESON_TAC[]);;

let WF_REC_CASES' = prove
 (`!(<<) clauses.
        WF((<<):A->A->bool) /\
        ALL (\(s,t). tailadmissible(<<) (\f a. T) s t) clauses
        ==> ?f:A->B. !x. f x = CASEWISE clauses f x`,
  REWRITE_TAC[WF_REC_CASES; tailadmissible]);;

let RECURSION_CASEWISE = prove
 (`!clauses.
   (?(<<). WF(<<) /\
           ALL (\(s:P->A,t). tailadmissible(<<) (\f a. T) s t) clauses) /\
   (!s t s' t' f x y. MEM (s,t) clauses /\ MEM (s',t') clauses
                      ==> (s x = s' y) ==> (t f x = t' f y))
   ==> ?f:A->B. ALL (\(s,t). !x. f (s x) = t f x) clauses`,
  REPEAT GEN_TAC THEN REWRITE_TAC[IMP_IMP; CONJ_ASSOC] THEN
  DISCH_THEN(CONJUNCTS_THEN2 MP_TAC ASSUME_TAC) THEN
  DISCH_THEN(CHOOSE_THEN (MP_TAC o MATCH_MP WF_REC_CASES')) THEN
  MATCH_MP_TAC MONO_EXISTS THEN REPEAT STRIP_TAC THEN
  ASM_REWRITE_TAC[] THEN ASM_MESON_TAC[CASEWISE_WORKS]);;

let RECURSION_CASEWISE_PAIRWISE = prove
 (`!clauses.
        (?(<<). WF (<<) /\
                ALL (\(s,t). tailadmissible(<<) (\f a. T) s t) clauses) /\
        ALL (\(s,t). !f x y. (s x = s y) ==> (t f x = t f y)) clauses /\
        PAIRWISE (\(s,t) (s',t'). !f x y. (s x = s' y) ==> (t f x = t' f y))
                 clauses
        ==> (?f. ALL (\(s,t). !x. f (s x) = t f x) clauses)`,
  let lemma = prove
   (`!P. (!x y. P x y ==> P y x)
         ==> !l. (!x y. MEM x l /\ MEM y l ==> P x y) <=>
                 ALL (\x. P x x) l /\ PAIRWISE P l`,
    REWRITE_TAC[GSYM IMP_IMP; RIGHT_FORALL_IMP_THM; GSYM ALL_MEM] THEN
    REPEAT GEN_TAC THEN DISCH_TAC THEN LIST_INDUCT_TAC THEN
    REWRITE_TAC[PAIRWISE; MEM; GSYM ALL_MEM] THEN ASM_MESON_TAC[])
  and paired_lambda = prove
   (`(\x. P x) = (\(a,b). P (a,b))`,
    REWRITE_TAC[FUN_EQ_THM; FORALL_PAIR_THM]) in
  let pth = REWRITE_RULE[FORALL_PAIR_THM; paired_lambda] (ISPEC
    `\(s,t) (s',t'). !c x:A y:A. (s x = s' y) ==> (t c x = t' c y)` lemma) in
  let cth = prove(lhand(concl pth),MESON_TAC[]) in
  REWRITE_TAC[GSYM(MATCH_MP pth cth); RIGHT_IMP_FORALL_THM] THEN
  REWRITE_TAC[RECURSION_CASEWISE]);;

let SUPERADMISSIBLE_T = prove
 (`superadmissible(<<) (\f x. T) s t <=> tailadmissible(<<) (\f x. T) s t`,
  REWRITE_TAC[superadmissible; admissible]);;

let RECURSION_SUPERADMISSIBLE = REWRITE_RULE[GSYM SUPERADMISSIBLE_T]
        RECURSION_CASEWISE_PAIRWISE;;

(* ------------------------------------------------------------------------- *)
(* Instantiate the casewise recursion theorem for existential claim.         *)
(* Also make a first attempt to simplify the distinctness clause. This may   *)
(* yield a theorem with just the wellfoundedness "?(<<)" assumption, or it   *)
(* may be that and an additional distinctness one.                           *)
(* ------------------------------------------------------------------------- *)

let instantiate_casewise_recursion =
  let EXPAND_PAIRED_ALL_CONV =
    let pth0,pth1 = (CONJ_PAIR o prove)
     (`(ALL (\(s,t). P s t) [a,b] <=> P a b) /\
       (ALL (\(s,t). P s t) (CONS (a,b) l) <=> P a b /\ ALL (\(s,t). P s t) l)`,
      REWRITE_TAC[ALL]) in
    let conv0 = REWR_CONV pth0 and conv1 = REWR_CONV pth1 in
    let rec conv tm =
      try conv0 tm with Failure _ ->
      let th = conv1 tm in CONV_RULE (funpow 2 RAND_CONV conv) th in
    conv
  and FORALL_PAIR_CONV =
    let rule = GEN_REWRITE_RULE RAND_CONV [GSYM FORALL_PAIR_THM] in
    let rec depair l t =
      match l with
        [v] -> REFL t
      | v::vs -> rule(BINDER_CONV (depair vs) t) in
    fun parm parms ->
      let p = mk_var("P",mk_fun_ty (type_of parm) bool_ty) in
      let tm = list_mk_forall(parms,mk_comb(p,parm)) in
      GEN p (SYM(depair parms tm))
  and SIMPLIFY_CASE_DISTINCTNESS_CLAUSES =
    let LSYM =
      GEN_ALL o CONV_RULE(LAND_CONV(ONCE_DEPTH_CONV SYM_CONV)) o SPEC_ALL
    and evensimps = prove
     (`((2 * m + 2 = 2 * n + 1) <=> F) /\
       ((2 * m + 1 = 2 * n + 2) <=> F) /\
       ((2 * m = 2 * n + 1) <=> F) /\
       ((2 * m + 1 = 2 * n) <=> F) /\
       ((2 * m = SUC(2 * n)) <=> F) /\
       ((SUC(2 * m) = 2 * n) <=> F)`,
      REWRITE_TAC[] THEN REPEAT CONJ_TAC THEN
      DISCH_THEN(MP_TAC o AP_TERM `EVEN`) THEN
      REWRITE_TAC[EVEN_MULT; EVEN_ADD; ARITH; EVEN]) in
    let allsimps = itlist (mk_rewrites false)
     [EQ_ADD_RCANCEL; EQ_ADD_LCANCEL;
      EQ_ADD_RCANCEL_0; EQ_ADD_LCANCEL_0;
      LSYM EQ_ADD_RCANCEL_0; LSYM EQ_ADD_LCANCEL_0;
      EQ_MULT_RCANCEL; EQ_MULT_LCANCEL;
      EQT_INTRO(SPEC_ALL EQ_REFL);
      ADD_EQ_0; LSYM ADD_EQ_0;
      MULT_EQ_0; LSYM MULT_EQ_0;
      MULT_EQ_1; LSYM MULT_EQ_1;
      ARITH_RULE `(m + n = 1) <=> (m = 1) /\ (n = 0) \/ (m = 0) /\ (n = 1)`;
      ARITH_RULE `(1 = m + n) <=> (m = 1) /\ (n = 0) \/ (m = 0) /\ (n = 1)`;
      evensimps; ARITH_EQ] []
    and [simp1; simp2; simp3] = map MATCH_MP (CONJUNCTS
      (TAUT
       `((a <=> F) /\ (b <=> b) ==> ((a ==> b) <=> T)) /\
        ((a <=> a') /\ (a' ==> (b <=> T)) ==> ((a ==> b) <=> T)) /\
        ((a <=> a') /\ (a' ==> (b <=> b')) ==> ((a ==> b) <=> (a' ==> b')))`))
    and false_tm = `F` and and_tm = `(/\)`
    and eq_refl = EQT_INTRO(SPEC_ALL EQ_REFL) in
    fun tm ->
      let net = itlist (net_of_thm false) allsimps (!basic_rectype_net) in
      let RECTYPE_ARITH_EQ_CONV =
        TOP_SWEEP_CONV(REWRITES_CONV net) THENC
        GEN_REWRITE_CONV DEPTH_CONV [AND_CLAUSES; OR_CLAUSES] in
      let SIMPLIFY_CASE_DISTINCTNESS_CLAUSE tm =
        let avs,bod = strip_forall tm in
        let ant,cons = dest_imp bod in
        let ath = RECTYPE_ARITH_EQ_CONV ant in
        let atm = rand(concl ath) in
        let bth = CONJ ath
          (if atm = false_tm then REFL cons
                    else DISCH atm
                          (PURE_REWRITE_CONV[eq_refl; ASSUME atm] cons)) in
        let cth = try simp1 bth with Failure _ -> try simp2 bth
                  with Failure _ -> simp3 bth in
        itlist MK_FORALL avs cth in
    (DEPTH_BINOP_CONV and_tm SIMPLIFY_CASE_DISTINCTNESS_CLAUSE THENC
     GEN_REWRITE_CONV DEPTH_CONV [FORALL_SIMP; AND_CLAUSES]) tm in
  let ELIM_LISTOPS_CONV =
    PURE_REWRITE_CONV[PAIRWISE; ALL; GSYM CONJ_ASSOC; AND_CLAUSES] THENC
    TOP_DEPTH_CONV GEN_BETA_CONV in
  fun def ->
    try PART_MATCH I EXISTS_REFL def with Failure _ ->
    let f,bod = dest_exists def in
    let cjs = conjuncts bod in
    let eqs = map (snd o strip_forall) cjs in
    let lefts,rights = unzip(map dest_eq eqs) in
    let arglists = map (snd o strip_comb) lefts in
    let parms0 = freesl(unions arglists) in
    let parms = if parms0 <> [] then parms0 else [genvar aty] in
    let parm = end_itlist (curry mk_pair) parms in
    let ss = map (fun a -> mk_gabs(parm,end_itlist (curry mk_pair) a)) arglists
    and ts = map (fun a -> mk_abs(f,mk_gabs(parm,a))) rights in
    let clauses = mk_flist(map2 (curry mk_pair) ss ts) in
    let th0 = UNDISCH(ISPEC clauses RECURSION_SUPERADMISSIBLE) in
    let th1 = EQ_MP (GEN_ALPHA_CONV f (concl th0)) th0 in
    let rewr_forall_th = REWR_CONV(FORALL_PAIR_CONV parm parms) in
    let th2 = CONV_RULE (BINDER_CONV
                  (LAND_CONV(GABS_CONV rewr_forall_th) THENC
                   EXPAND_PAIRED_ALL_CONV)) th1 in
    let f2,bod2 = dest_exists(concl th2) in
    let ths3 = map
     (CONV_RULE (COMB2_CONV (funpow 2 RAND_CONV GEN_BETA_CONV)
                (RATOR_CONV BETA_CONV THENC GEN_BETA_CONV)) o SPEC_ALL)
     (CONJUNCTS(ASSUME bod2)) in
    let ths4 = map2
     (fun th t -> let avs,tbod = strip_forall t in
                  itlist GEN avs (PART_MATCH I th tbod)) ths3 cjs in
    let th5 = SIMPLE_EXISTS f (end_itlist CONJ ths4) in
    let th6 = PROVE_HYP th2 (SIMPLE_CHOOSE f th5) in
    let th7 =
     (RAND_CONV(COMB2_CONV
          (RAND_CONV (LAND_CONV (GABS_CONV(BINDER_CONV
                   (BINDER_CONV(rewr_forall_th) THENC rewr_forall_th)))))
          (LAND_CONV (funpow 2 GABS_CONV(BINDER_CONV
                   (BINDER_CONV(rewr_forall_th) THENC rewr_forall_th))))) THENC
      ELIM_LISTOPS_CONV) (hd(hyp th6)) in
    let th8 = PROVE_HYP (UNDISCH(snd(EQ_IMP_RULE th7))) th6 in
    let wfasm,cdasm = dest_conj(hd(hyp th8)) in
    let th9 = PROVE_HYP (CONJ (ASSUME wfasm) (ASSUME cdasm)) th8 in
    let th10 = SIMPLIFY_CASE_DISTINCTNESS_CLAUSES cdasm in
    let th11 = PROVE_HYP (UNDISCH(snd(EQ_IMP_RULE th10))) th9 in
    PROVE_HYP TRUTH th11;;

(* ------------------------------------------------------------------------- *)
(* Justify existence assertion and try to simplify/remove side-conditions.   *)
(* ------------------------------------------------------------------------- *)

let pure_prove_recursive_function_exists =
  let tuple_function_existence tm =
    let f,def = dest_exists tm in
    let domtys0,ranty0 = splitlist dest_fun_ty (type_of f) in
    let nargs =
      itlist
       (max o length o snd o strip_comb o lhs o snd o strip_forall)
       (conjuncts(snd(strip_forall def))) 0 in
    let domtys,midtys = chop_list nargs domtys0 in
    let ranty = itlist mk_fun_ty midtys ranty0 in
    if length domtys <= 1 then ASSUME tm else
    let dty = end_itlist (fun ty1 ty2 -> mk_type("prod",[ty1;ty2])) domtys in
    let f' = variant (frees tm) (mk_var(fst(dest_var f),mk_fun_ty dty ranty)) in
    let gvs = map genvar domtys in
    let f'' = list_mk_abs(gvs,mk_comb(f',end_itlist (curry mk_pair) gvs)) in
    let def' = subst [f'',f] def in
    let th1 = EXISTS (tm,f'') (ASSUME def')
    and bth = BETAS_CONV (list_mk_comb(f'',gvs)) in
    let th2 = GEN_REWRITE_CONV TOP_DEPTH_CONV [bth] (hd(hyp th1)) in
    SIMPLE_CHOOSE f' (PROVE_HYP (UNDISCH(snd(EQ_IMP_RULE th2))) th1) in
  let is_applicative =
    let rec applicative f tm =
      match tm with
        Comb(t,x) -> (t = f or applicative f t) & applicative f x
      | Abs(x,t) -> applicative f t
      | Var(_,_) -> tm <> f
      | Const(_,_,_) -> true in
    fun tm -> let f,bod = dest_abs tm in
              let vs,t = dest_gabs bod in
              free_in f vs or applicative f t in
  let LAMBDA_PAIR_CONV =
    let rewr1 =  GEN_REWRITE_RULE I [GSYM FORALL_PAIR_THM]
    and rewr2 = GEN_REWRITE_CONV I [FUN_EQ_THM] in
    fun parms tm ->
      let parm = end_itlist (curry mk_pair) parms in
      let x,bod = dest_abs tm in
      let tm' = mk_gabs(parm,vsubst[parm,x] bod) in
      let th1 = BETA_CONV(mk_comb(tm,parm))
      and th2 = GEN_BETA_CONV (mk_comb(tm',parm)) in
      let th3 = TRANS th1 (SYM th2) in
      let th4 = itlist (fun v th -> rewr1 (GEN v th))
                       (butlast parms) (GEN (last parms) th3) in
      EQ_MP (SYM(rewr2(mk_eq(tm,tm')))) th4 in
  let break_down_admissibility th1 =
    if hyp th1 = [] then th1 else
    let def = concl th1 in
    let f,bod = dest_exists def in
    let cjs = conjuncts bod in
    let eqs = map (snd o strip_forall) cjs in
    let lefts,rights = unzip(map dest_eq eqs) in
    let arglists = map (snd o strip_comb) lefts in
    let parms0 = freesl(unions arglists) in
    let parms = if parms0 <> [] then parms0 else [genvar aty] in
    let parm = end_itlist (curry mk_pair) parms in
    let wfasm = find is_exists (hyp th1) in
    let ord,bod = dest_exists wfasm in
    let LAMBDA_CONV = LAMBDA_PAIR_CONV parms in
    let INSTANTIATE_TYPE_RULE =
      let atype = type_of(rand(hd lefts))
      and btype = type_of(hd rights)
      and ptype = type_of parm in
      INST_TYPE[atype,mk_vartype "A"; btype,mk_vartype "B";
                ptype,mk_vartype "P"] in
    let PSPEC th =
      let vname,vty = dest_var(fst(dest_forall(concl th))) in
      let fty,xyty = dest_fun_ty vty in
      let ranty = snd(dest_fun_ty xyty) in
      let ty' = itlist (mk_fun_ty o type_of) parms ranty in
      let v = mk_var(vname,mk_fun_ty fty ty') in
      let f = genvar fty in
      BETA_RULE(SPEC(mk_abs(f,mk_gabs(parm,list_mk_comb(v,f::parms)))) th)
    and QSPEC th =
      let vname,vty = dest_var(fst(dest_forall(concl th))) in
      let ranty = snd(dest_fun_ty vty) in
      let ty' = itlist (mk_fun_ty o type_of) parms ranty in
      let v = mk_var(vname,ty') in
      BETA_RULE(SPEC(mk_gabs(parm,list_mk_comb(v,parms))) th)
    and RSPEC th =
      let vname,vty = dest_var(fst(dest_forall(concl th))) in
      let uty,bty = dest_fun_ty vty in
      let fty,xyty = dest_fun_ty bty in
      let ranty = snd(dest_fun_ty xyty) in
      let ty' = itlist (mk_fun_ty o type_of) parms ranty in
      let v = mk_var(vname,mk_fun_ty uty (mk_fun_ty fty ty')) in
      let u = genvar uty and f = genvar fty in
      BETA_RULE(SPEC(list_mk_abs
       ([u;f],mk_gabs(parm,list_mk_comb(v,u::f::parms)))) th)
    and VSPEC th = snd(SPEC_VAR th) in
    let HACK_PROFORMA vs =
      CONV_RULE(TOP_DEPTH_CONV GEN_BETA_CONV) o
      CONV_RULE(ONCE_DEPTH_CONV(ABS_CONV LAMBDA_CONV)) o
      itlist I vs o
      INSTANTIATE_TYPE_RULE in
    let SUPERADMISSIBLE_COND' =
      HACK_PROFORMA [PSPEC; PSPEC; VSPEC; PSPEC; VSPEC; VSPEC]
                    SUPERADMISSIBLE_COND
    and SUPERADMISSIBLE_TAIL' =
      HACK_PROFORMA [PSPEC; VSPEC; VSPEC; VSPEC] TAIL_IMP_SUPERADMISSIBLE
    and ADMISSIBLE_COND' =
      HACK_PROFORMA [PSPEC; PSPEC; VSPEC; PSPEC; VSPEC; VSPEC]
                    ADMISSIBLE_COND
    and ADMISSIBLE_COMB' =
      HACK_PROFORMA [PSPEC; PSPEC; VSPEC; VSPEC; VSPEC] ADMISSIBLE_COMB
    and ADMISSIBLE_BASE' =
      HACK_PROFORMA [QSPEC; VSPEC; VSPEC; VSPEC] ADMISSIBLE_BASE
    and ADMISSIBLE_NEST' =
      HACK_PROFORMA [PSPEC; VSPEC; VSPEC; VSPEC] ADMISSIBLE_NEST
    and ADMISSIBLE_LAMBDA' =
      HACK_PROFORMA [RSPEC; QSPEC; VSPEC; VSPEC] ADMISSIBLE_LAMBDA in
    let ADMISS_TAC =
      CONJ_TAC ORELSE
      MATCH_MP_TAC SUPERADMISSIBLE_COND' ORELSE
      MATCH_MP_TAC SUPERADMISSIBLE_TAIL' ORELSE
      MATCH_MP_TAC ADMISSIBLE_IMP_SUPERADMISSIBLE ORELSE
      MATCH_ACCEPT_TAC ADMISSIBLE_CONST ORELSE
      (MATCH_MP_TAC ADMISSIBLE_SUM THEN GEN_TAC THEN BETA_TAC) ORELSE
      (MATCH_MP_TAC ADMISSIBLE_NSUM THEN GEN_TAC THEN BETA_TAC) ORELSE
      MATCH_MP_TAC ADMISSIBLE_COND' ORELSE
      MATCH_MP_TAC ADMISSIBLE_BASE' ORELSE
      MATCH_MP_TAC ADMISSIBLE_NEST' ORELSE
      (MATCH_MP_TAC ADMISSIBLE_LAMBDA' THEN GEN_TAC) ORELSE
      (let tac = MATCH_MP_TAC ADMISSIBLE_COMB' in
       W(fun (_,w) -> if is_applicative(rand w) then tac else NO_TAC)) in
    let SIMP_ADMISS_TAC =
      REWRITE_TAC[LET_DEF; LET_END_DEF] THEN
      REPEAT ADMISS_TAC THEN
      TRY(W(fun (asl,w) -> let v = fst(dest_forall w) in
              X_GEN_TAC v THEN
              MAP_EVERY
                (fun v -> TRY(GEN_REWRITE_TAC I [FORALL_PAIR_THM]) THEN
                          X_GEN_TAC v)
                parms THEN
              CONV_TAC(TOP_DEPTH_CONV GEN_BETA_CONV) THEN
              MAP_EVERY (fun v -> SPEC_TAC(v,v)) (rev parms @ [v]))) THEN
      PURE_REWRITE_TAC[FORALL_SIMP] THEN
      W(fun (asl,w) -> MAP_EVERY (fun t -> SPEC_TAC(t,t))
                                 (subtract (frees w) [ord])) THEN
      W(fun (asl,w) -> ACCEPT_TAC(ASSUME w)) in
    let th2 = prove(bod,SIMP_ADMISS_TAC) in
    let th3 = SIMPLE_EXISTS ord th2 in
    let allasms = hyp th3 and wfasm = lhand(concl th2) in
    let th4 = ASSUME(list_mk_conj(wfasm::subtract allasms [wfasm])) in
    let th5 = SIMPLE_CHOOSE ord (itlist PROVE_HYP (CONJUNCTS th4) th3) in
    PROVE_HYP th5 th1 in
  fun etm ->
    let eth = tuple_function_existence etm in
    let dtm = hd(hyp eth) in
    let th0 = break_down_admissibility(instantiate_casewise_recursion dtm) in
    let th = PROVE_HYP th0 eth in
    if concl th = etm then th
    else failwith "prove_general_recursive_function_exists: sanity";;

(* ------------------------------------------------------------------------- *)
(* Same, but attempt to prove the wellfoundedness hyp by good guesses.       *)
(* ------------------------------------------------------------------------- *)

let prove_general_recursive_function_exists =
  let prove_depth_measure_exists =
    let num_ty = `:num` in
    fun tyname ->
      let _,_,sth = assoc tyname (!inductive_type_store) in
      let ty,zty = dest_fun_ty
       (type_of(fst(dest_exists(snd(strip_forall(concl sth)))))) in
      let rth = INST_TYPE [num_ty,zty] sth in
      let avs,bod = strip_forall(concl rth) in
      let ev,cbod = dest_exists bod in
      let process_clause k t =
        let avs,eq = strip_forall t in
        let l,r = dest_eq eq in
        let fn,cargs = dest_comb l in
        let con,args = strip_comb cargs in
        let bargs = filter (fun t -> type_of t = ty) args in
        let r' = list_mk_binop `(+):num->num->num`
                  (mk_small_numeral k :: map (curry mk_comb fn) bargs) in
        list_mk_forall(avs,mk_eq(l,r')) in
      let cjs = conjuncts cbod in
      let def = map2 process_clause (1--length cjs) cjs in
      prove_recursive_functions_exist sth (list_mk_conj def) in
  let INDUCTIVE_MEASURE_THEN tac (asl,w) =
    let ev,bod = dest_exists w in
    let ty = fst(dest_type(fst(dest_fun_ty(type_of ev)))) in
    let th = prove_depth_measure_exists ty in
    let ev',bod' = dest_exists(concl th) in
    let th' = INST_TYPE(type_match (type_of ev') (type_of ev) []) th in
    (MP_TAC th' THEN MATCH_MP_TAC MONO_EXISTS THEN
     GEN_TAC THEN DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN tac) (asl,w) in
  let CONSTANT_MEASURE_THEN =
    let one_tm = `1` in
    fun tac (asl,w) ->
      let ev,bod = dest_exists w in
      let ty = fst(dest_fun_ty(type_of ev)) in
      (EXISTS_TAC(mk_abs(genvar ty,one_tm)) THEN tac) (asl,w) in
  let GUESS_MEASURE_THEN tac =
    (EXISTS_TAC `\n. n + 1` THEN tac) ORELSE
    (INDUCTIVE_MEASURE_THEN tac) ORELSE
    CONSTANT_MEASURE_THEN tac in
  let pth_lexleft = prove
   (`(?r. WF(r) /\
          ?s. WF(s) /\
              P(\(x1,y1) (x2,y2). r x1 x2 \/ (x1 = x2) /\ s y1 y2))
     ==> ?t:A#B->A#B->bool. WF(t) /\ P t`,
    REPEAT STRIP_TAC THEN EXISTS_TAC
     `\(x1:A,y1:B) (x2:A,y2:B). r x1 x2 \/ (x1 = x2) /\ s y1 y2` THEN
    ASM_SIMP_TAC[WF_LEX]) in
  let pth_lexright = prove
   (`(?r. WF(r) /\
          ?s. WF(s) /\
              P(\(x1,y1) (x2,y2). r y1 y2 \/ (y1 = y2) /\ s x1 x2))
     ==> ?t:A#B->A#B->bool. WF(t) /\ P t`,
    REPEAT STRIP_TAC THEN
    EXISTS_TAC `\u:A#B v:A#B.
                  (\(x1:B,y1:A) (x2:B,y2:A). r x1 x2 \/ (x1 = x2) /\ s y1 y2)
                   ((\(a,b). b,a) u) ((\(a,b). b,a) v)` THEN
    ASM_SIMP_TAC[ISPEC `\(a,b). b,a` WF_MEASURE_GEN; WF_LEX; ETA_AX] THEN
    FIRST_X_ASSUM(fun th -> MP_TAC th THEN
                            MATCH_MP_TAC(TAUT `(a = b) ==> a ==> b`) THEN
                            AP_TERM_TAC) THEN
    REWRITE_TAC[FUN_EQ_THM; FORALL_PAIR_THM]) in
  let pth_measure = prove
   (`(?m:A->num. P(measure m)) ==> ?r:A->A->bool. WF(r) /\ P r`,
    MESON_TAC[WF_MEASURE]) in
  let rec GUESS_WF_THEN tac (asl,w) =
   ((MATCH_MP_TAC pth_lexleft THEN GUESS_WF_THEN (GUESS_WF_THEN tac)) ORELSE
    (MATCH_MP_TAC pth_lexright THEN GUESS_WF_THEN (GUESS_WF_THEN tac)) ORELSE
    (MATCH_MP_TAC pth_measure THEN
     REWRITE_TAC[measure; MEASURE_LE] THEN
     REWRITE_TAC[FORALL_PAIR_THM] THEN
     GUESS_MEASURE_THEN tac)) (asl,w) in
  let GUESS_ORDERING_TAC =
    let false_tm = `\x:A y:A. F` in
    W(fun (asl,w) -> let ty = fst(dest_fun_ty(type_of(fst(dest_exists w)))) in
                     EXISTS_TAC(inst [ty,aty] false_tm) THEN
                     REWRITE_TAC[WF_FALSE] THEN NO_TAC) ORELSE
    GUESS_WF_THEN
     (REWRITE_TAC[FORALL_PAIR_THM] THEN ARITH_TAC) in
  fun etm ->
    let th = pure_prove_recursive_function_exists etm in
    try let wtm = find is_exists (hyp th) in
        let wth = prove(wtm,GUESS_ORDERING_TAC) in
        PROVE_HYP wth th
    with Failure _ -> th;;

(* ------------------------------------------------------------------------- *)
(* Simple "define" function.                                                 *)
(* ------------------------------------------------------------------------- *)

let define =
  let close_definition_clauses tm =
    let avs,bod = strip_forall tm in
    let cjs = conjuncts bod in
    let fs =
      try map (repeat rator o lhs o snd o strip_forall) cjs
      with Failure _ -> failwith "close_definition_clauses: non-equation" in
    if length (setify fs) <> 1
    then failwith "close_definition_clauses: defining multiple functions" else
    let f = hd fs in
    if mem f avs then failwith "close_definition_clauses: fn quantified" else
    let do_clause t =
      let lvs,bod = strip_forall t in
      let fvs = subtract (frees(lhs bod)) (f::lvs) in
      SPECL fvs (ASSUME(list_mk_forall(fvs,t))) in
    let ths = map do_clause cjs in
    let ajs = map (hd o hyp) ths in
    let th = ASSUME(list_mk_conj ajs) in
    f,itlist GEN avs (itlist PROVE_HYP (CONJUNCTS th) (end_itlist CONJ ths)) in
  fun tm ->
    let tm' = snd(strip_forall tm) in
    try let th,th' = tryfind (fun th -> th,PART_MATCH I th tm')
                             (!the_definitions) in
        if can (PART_MATCH I th') (concl th) then
         (warn true "Benign redefinition"; th')
        else failwith ""
    with Failure _ ->
      let f,th = close_definition_clauses tm in
      let etm = mk_exists(f,hd(hyp th)) in
      let th1 = prove_general_recursive_function_exists etm in
      let th2 = new_specification[fst(dest_var f)] th1 in
      let g = mk_mconst(dest_var f) in
      let th3 = PROVE_HYP th2 (INST [g,f] th) in
      the_definitions := th3::(!the_definitions); th3;;
